(*
 * Hoare Logic,
 * Annotated Examples
 * 
 * CVS : $Id: tlhlAnnotExamples.v,v 1.8 2005/05/13 19:28:12 amanous Exp $
*)

Require Export Sumbool.
Require Export tlhlAnnotWP.
Require Export Classical_Prop.



Definition gInt : Env := fun var : Var => uint.


(*******************************************************************)
Section ExampleSwap.
(*******************************************************************)

Let x : Var := 1%nat.
Let y : Var := 2%nat.
Let z : Var := 3%nat.

Let xx : Var := 4%nat.
Let yy : Var := 5%nat.

Definition pSwap :=
   acseq (acseq (acassign z (aevar x))
     (acassign x (aevar y)))
   (acassign y (aevar z))
.

Definition dSwap :
   typAC gInt pSwap
.
Proof.
typeAC.
Defined.

Lemma wpSwap :
  wpC gInt pSwap dSwap (fun s : Store gInt => s y = s xx /\ s x = s yy)
    = (fun s : Store gInt => s x = s xx /\ s y = s yy)
.
Proof.
reflexivity.
Qed.

Theorem sSwap :
   SpecAC gInt pSwap
     (fun s : Store gInt => s x = s xx /\ s y = s yy)
     (fun s : Store gInt => s y = s xx /\ s x = s yy)
.
Proof.
specAC dSwap.
Qed.

End ExampleSwap.


(*******************************************************************)
Section ExampleInc.
(*******************************************************************)

Let x  : Var := 1%nat.
Let xx : Var := 2%nat.

Lemma sIncpre :
   pred_implies gInt
      (fun s : Store gInt => s x = s xx)
      (fun s : Store gInt => s x + 1 = s xx + 1)
.
Proof.
unfold pred_implies in |- *.
intros s.
intros H.
apply Zplus_eq_compat with
    (n := s x)
    (m := s xx)
    (p := 1)
    (q := 1).
 assumption.
 trivial.
Qed.

Definition pInc :=
   acseq
      (acassert gInt
         (fun s : Store gInt => s x = s xx)
         (fun s : Store gInt => s x + 1 = s xx + 1)
         sIncpre)
      (acassign x (aebinop opPlus (aevar x) (aeint 1)))
.

Theorem dInc :
   typAC gInt pInc
.
Proof.
typeAC.
Defined.

Lemma wpInc :
  wpC gInt pInc dInc (fun s : Store gInt => s x = s xx + 1)
    = (fun s : Store gInt => s x = s xx)
.
Proof.
reflexivity.
Qed.

Theorem sInc :
   SpecAC gInt pInc
     (fun s : Store gInt => s x = s xx)
     (fun s : Store gInt => s x = s xx + 1)
.
Proof.
specAC dInc.
Qed.

End ExampleInc.


(*******************************************************************)
Section ExampleMax.
(*******************************************************************)

Let a : Var := 1%nat.
Let b : Var := 2%nat.
Let z : Var := 3%nat.

Lemma pMax0 :
  pred_implies gInt
  (fun s : Store gInt => True)
  (fun s : Store gInt =>
    (if Zgt_bool (s a) (s b)
      then (fun s : Store gInt => s a > s b)
      else (fun s : Store gInt => ~(s a > s b))) s)
.
Proof.
unfold pred_implies in |- *.
intros s H.
destruct (sumbool_of_bool (Zgt_bool (s a) (s b))).
 rewrite e.
   unfold Zgt in |- *.
   unfold Zgt_bool in e.
   destruct (s a ?= s b).
  discriminate e.
  discriminate e.
  reflexivity.
 rewrite e.
   unfold Zgt in |- *.
   unfold Zgt_bool in e.
   destruct (s a ?= s b).
  discriminate.
  discriminate.
  discriminate e.
Qed.

Lemma pMax1 :
  pred_implies gInt
    (fun s : Store gInt => s a > s b)
    (fun s : Store gInt => s a >= s a /\ s a >= s b)
.
Proof.
unfold pred_implies in |- *.
intros s H.
constructor.
 apply Zle_ge.
   apply Zle_refl.
 unfold Zge in |- *.
   unfold Zgt_bool in H.
   unfold Zgt in H.
   destruct (s a ?= s b).
  discriminate.
  discriminate H.
  discriminate.
Qed.

Lemma pMax2 :
  pred_implies gInt
    (fun s : Store gInt => ~(s a > s b))
    (fun s : Store gInt => s b >= s a /\ s b >= s b)
.
Proof.
unfold pred_implies in |- *.
intros s H.
constructor.
 apply Zle_ge.
   unfold Zle in |- *.
   unfold Zgt in H.
   destruct (s a ?= s b).
  discriminate.
  discriminate.
  assumption.
 apply Zle_ge.
   apply Zle_refl.
Qed.

Definition pMax :=
  acseq
    (acassert gInt
      (fun s : Store gInt => True)
      (fun s : Store gInt =>
        (if Zgt_bool (s a) (s b)
          then (fun s : Store gInt => s a > s b)
          else (fun s : Store gInt => ~(s a > s b))) s)
      pMax0)
    (acif (aebinop opGt (aevar a) (aevar b))
      (acseq
        (acassert gInt
          (fun s : Store gInt => s a > s b)
          (fun s : Store gInt => s a >= s a /\ s a >= s b)
          pMax1)
        (acassign z (aevar a)))
      (acseq
        (acassert gInt
          (fun s : Store gInt => ~(s a > s b))
          (fun s : Store gInt => s b >= s a /\ s b >= s b)
          pMax2)
        (acassign z (aevar b))))
.

Theorem dMax :
  typAC gInt pMax
.
Proof.
typeAC.
Defined.

Lemma wpMax :
   wpC gInt pMax dMax (fun s : Store gInt => s z >= s a /\ s z >= s b) =
   (fun s : Store gInt => True)
.
Proof.
reflexivity.
Qed.

Theorem sMax :
  SpecAC gInt pMax
    (fun s : Store gInt => True)
    (fun s : Store gInt => s z >= s a /\ s z >= s b)
.
Proof.
specAC dMax.
Qed.

End ExampleMax.


(*******************************************************************)
Section ExampleLog2.
(*******************************************************************)

Let xx : Var := 0%nat.
Let n  : Var := 1%nat.
Let m  : Var := 2%nat.

Lemma pLog2pre :
   pred_implies gInt
      (fun s : Store gInt => s n >= 1 /\ s n = s xx)
      (fun s : Store gInt => s xx / 2 ^ 0 = s n /\ s n >= 1 /\ 0 >= 0)
.
Proof.
unfold pred_implies in |- *.
intros s H.
destruct H as (H1, H2).
constructor.
 rewrite H2.
   apply Zdiv_by_1.
 constructor.
  assumption.
  auto with zarith.
Qed.

Lemma pLog2cond :
   pred_implies gInt
      (fun s : Store gInt => s xx / 2 ^ s m = s n /\ s n >= 1 /\ s m >= 0)
      (fun s : Store gInt => (if Zgt_bool (s n) 1 then
                                 (fun s : Store gInt =>
                                    s xx / 2 ^ s m = s n /\ s n >= 1 /\
                                    s m >= 0 /\ s n > 1)
                              else
                                 (fun s : Store gInt =>
                                    s xx / 2 ^ s m = s n /\ s n >= 1 /\
                                    s m >= 0 /\ s n <= 1)) s)
.
Proof.
unfold pred_implies in |- *.
intros s H.
destruct H as (H1, (H2, H3)).
destruct (sumbool_of_bool (Zgt_bool (s n) 1)).
 rewrite e.
   constructor.
  assumption.
  constructor.
   assumption.
   constructor.
    assumption.
    unfold Zgt in |- *.
      unfold Zgt_bool in e.
      induction (s n ?= 1).
     discriminate e.
     discriminate e.
     reflexivity.
 rewrite e.
   constructor.
  assumption.
  constructor.
   assumption.
   constructor.
    assumption.
    unfold Zle in |- *.
      unfold not in |- *.
      unfold Zgt_bool in e.
      intro H.
      induction (s n ?= 1).
     discriminate H.
     discriminate H.
     discriminate e.
Qed.

Lemma pLog2while :
   pred_implies gInt
      (fun s : Store gInt => s xx / 2 ^ s m = s n /\ s n >= 1 /\ s m >= 0 /\
                             s n > 1)
      (fun s : Store gInt => s xx / 2 ^ (s m + 1) = s n / 2 /\ s n / 2 >= 1 /\
                             s m + 1 >= 0)
.
Proof.
unfold pred_implies in |- *.
intros s H.
destruct H as (H1, (H2, (H3, H4))).
constructor.
 rewrite <- H1.
   transitivity (s xx / (2 ^ s m * 2)).
  assert (2 ^ (s m + 1) = 2 ^ s m * 2).
   change (2 ^ (s m + 1) = 2 ^ s m * 2 ^ 1) in |- *.
     apply Zpower_exp.
    assumption.
    auto with zarith.
   rewrite H.
     reflexivity.
  apply Zmult_div_assoc.
   apply Zpow_gt_zero.
    auto with zarith.
    assumption.
   auto with zarith.
 constructor.
  change (s n / 2 >= 2 / 2) in |- *.
    apply Z_div_ge.
   auto with zarith.
   apply Zle_ge.
     change (Zsucc 1 <= s n) in |- *.
     apply Zlt_le_succ.
     apply Zgt_lt.
     assumption.
  apply Zge_trans with (m := s m).
   auto with zarith.
   assumption.
Qed.

Lemma pLog2post :
   pred_implies gInt
      (fun s : Store gInt => s xx / 2 ^ s m = s n /\ s n >= 1 /\ s m >= 0 /\
                             s n <= 1)
      (fun s : Store gInt => 2 ^ s m <= s xx < 2 ^ (s m + 1))
.
Proof.
unfold pred_implies in |- *.
intros s H.
destruct H as (H1, (H2, (H3, H4))).
apply Zdiv_pow2_1.
 assumption.
 assert (s n = 1).
  apply Zle_antisym.
   assumption.
   apply Zge_le.
     assumption.
  rewrite <- H.
    assumption.
Qed.

Definition pLog2 :=
  acseq
   (acseq
     (acseq
      (acassert gInt
         (fun s : Store gInt => s n >= 1 /\ s n = s xx)
         (fun s : Store gInt => s xx / 2 ^ 0 = s n /\ s n >= 1 /\ 0 >= 0)
         pLog2pre)
      (acassign m (aeint 0)))
     (acwhile gInt
       (fun s : Store gInt => s xx / 2 ^ s m = s n /\ s n >= 1 /\ s m >= 0)
       (aeweak gInt
         (fun s : Store gInt => s xx / 2 ^ s m = s n /\ s n >= 1 /\ s m >= 0)
         (fun s : Store gInt => (if Zgt_bool (s n) 1 then
                                   (fun s : Store gInt =>
                                     s xx / 2 ^ s m = s n /\ s n >= 1 /\
                                     s m >= 0 /\ s n > 1)
                                 else
                                   (fun s : Store gInt =>
                                     s xx / 2 ^ s m = s n /\ s n >= 1 /\
                                     s m >= 0 /\ s n <= 1)) s)
         pLog2cond (aebinop opGt (aevar n) (aeint 1)))
       (acseq
         (acassert gInt
            (fun s : Store gInt => s xx / 2 ^ s m = s n /\ s n >= 1 /\
                                   s m >= 0 /\
                                   s n > 1)
            (fun s : Store gInt => s xx / 2 ^ (s m + 1) = s n / 2 /\
                                   s n / 2 >= 1 /\ s m + 1 >= 0)
            pLog2while)
            (acseq
               (acassign n (aebinop opDiv (aevar n) (aeint 2)))
               (acassign m (aebinop opPlus (aevar m) (aeint 1)))))))
   (acassert gInt
     (fun s : Store gInt => s xx / 2 ^ s m = s n /\ s n >= 1 /\ s m >= 0 /\
                            s n <= 1)
     (fun s : Store gInt => 2 ^ s m <= s xx < 2 ^ (s m + 1))
     pLog2post)
.

Theorem dLog2 :
   typAC gInt pLog2
.
Proof.
typeAC.
Defined.

Lemma wpLog2 :
   wpC gInt pLog2 dLog2
     (fun s : Store gInt => 2 ^ (s m) <= s xx < 2 ^ (s m + 1))
   = (fun s : Store gInt => s n >= 1 /\ s n = s xx)
.
Proof.
reflexivity.
Qed.

Theorem sLog2 :
   SpecAC gInt pLog2
     (fun s : Store gInt => s n >= 1 /\ s n = s xx)
     (fun s : Store gInt => 2 ^ (s m) <= s xx < 2 ^ (s m + 1)).
Proof.
specAC dLog2.
Qed.

End ExampleLog2.

(*******************************************************************)
Section ExampleFactorial.
(*******************************************************************)

Let x : Var := 0%nat.
Let n : Var := 1%nat.
Let r : Var := 2%nat.

Lemma pFactpre :
   pred_implies gInt
      (fun s : Store gInt => s n >= 0)
      (fun s : Store gInt => 1 = Zfact 0 /\ 0 <= s n /\ 0 >= 0).

Proof.
unfold pred_implies.
intros s H.
constructor.
   trivial.
constructor.
   apply Zge_le.
   assumption.
   auto with zarith.
Qed.

Lemma pFactcond :
   pred_implies gInt
      (fun s : Store gInt => s r = Zfact (s x) /\ s x <= s n /\ s x >= 0)
      (fun s : Store gInt => (if Zlt_bool (s x) (s n) then
                                 (fun s : Store gInt => s r = Zfact (s x) /\
                                                        s x <= s n /\
                                                        s x >= 0 /\
                                                        s x < s n)
                              else
                                 (fun s : Store gInt => s r = Zfact (s x) /\
                                                        s x <= s n /\
                                                        s x >= 0 /\
                                                        s x >= s n)) s)
.

Proof.
unfold pred_implies.
intros s H.
destruct H as (H1, (H2, H3)).
destruct (sumbool_of_bool (Zlt_bool (s x) (s n))).
rewrite e.
constructor.
assumption.
constructor.
assumption.
constructor.
assumption.
unfold Zlt in |- *.
unfold Zlt_bool in e.
induction (s x ?= s n).
discriminate e.
trivial.
discriminate e.
rewrite e.
constructor.
assumption.
constructor.
assumption.
constructor.
assumption.
unfold Zge in |- *.
unfold Zlt_bool in e.
induction (s x ?= s n).
discriminate.
discriminate.
discriminate.
Qed.
 
Lemma pFactwhile :
   pred_implies gInt
      (fun s : Store gInt => s r = Zfact (s x) /\ s x <= s n /\ s x >= 0 /\
                             s x < s n)
      (fun s : Store gInt => (s r) * (s x + 1) = Zfact (s x + 1) /\
                             s x + 1 <= s n /\ s x + 1 >= 0)
.
  
Proof.
unfold pred_implies.
intros s H.
destruct H as (H1, (H2, (H3, H4))).
constructor.
   rewrite H1.
   apply sym_eq.
   apply ZfactS with (n := s x). 
   assumption.
   constructor.
   change (s x + 1) with ( Zsucc (s x)).
   apply Zlt_le_succ.
   assumption.
   change (s x + 1) with ( Zsucc (s x)).
   auto with zarith.
Qed.

Lemma pFactpost :
   pred_implies gInt
      (fun s : Store gInt => s r = Zfact (s x) /\ s x <= s n /\ s x >= 0 /\
                             s x >= s n)
      (fun s: Store gInt => s r = Zfact (s n))
.

Proof.
unfold pred_implies.
intros s H.
destruct H as (H1, (H2, (H3, H4))).
assert (s x = s n).
apply Zle_antisym.
assumption.
apply Zge_le.
assumption.
rewrite <- H.
assumption.
Qed.

Definition pFact :=
   acseq
      (acseq
         (acassert gInt
            (fun s : Store gInt => s n >= 0)
            (fun s : Store gInt => 1 = Zfact 0 /\ 0 <= s n /\ 0 >= 0)
            pFactpre
         )
         (acseq
            (acassign x (aeint 0)) 
            (acassign r (aeint 1))
         )
      )
      (acseq
         (acwhile gInt
            (fun s : Store gInt => s r = Zfact (s x) /\ s x <= s n /\ s x >= 0)
            (aeweak gInt
               (fun s : Store gInt => s r = Zfact (s x) /\ s x <= s n /\
                                      s x >= 0)
               (fun s : Store gInt => (if Zlt_bool (s x) (s n) then
                                         (fun s : Store gInt =>
                                            s r = Zfact (s x) /\
                                            s x <= s n /\ s x >= 0 /\
                                            s x < s n)
                                       else
                                         (fun s : Store gInt =>
                                            s r = Zfact (s x) /\
                                            s x <= s n /\ s x >= 0 /\
                                            s x >= s n)) s)
               pFactcond (aebinop opLt (aevar x) (aevar n)))
            (acseq
               (acassert gInt
                  (fun s : Store gInt => s r = Zfact (s x) /\ s x <= s n /\
                                         s x >= 0 /\ s x < s n)
                  (fun s : Store gInt => (s r) * (s x + 1) = Zfact (s x + 1) /\
                                         s x +1 <= s n /\ s x + 1 >= 0)
                  pFactwhile
               )
               (acseq
                  (acassign x (aebinop opPlus (aevar x) (aeint 1)))
                  (acassign r (aebinop opTimes (aevar r) (aevar x)))
               )
            )
         )
         (acassert gInt
            (fun s : Store gInt => s r = Zfact (s x) /\ s x <= s n /\
                                   s x >= 0 /\ s x >= s n)
            (fun s: Store gInt => s r = Zfact (s n))
            pFactpost
         )
      )
.

Theorem dFact :
   typAC gInt pFact.

Proof.
typeAC.
Defined.


Lemma wpFact :
   wpC gInt pFact dFact
      (fun s: Store gInt => s r = Zfact (s n))
      = (fun s: Store gInt => s n >= 0).

Proof.
reflexivity.
Qed.


Theorem sFact :
   SpecAC gInt pFact
      (fun s: Store gInt => s n >= 0 )
      (fun s: Store gInt => s r = Zfact (s n)).

Proof.
specAC dFact.
Qed.

End ExampleFactorial.

(*******************************************************************)
Section ExamplePrimality.
(*******************************************************************)

Let n  : Var := 0%nat.
Let i  : Var := 1%nat.
Let pr : Var := 2%nat.

Definition gPr : Env :=
   fun v : Var =>
   match v with
      2%nat => ubool
      | _ => uint
   end
.

(* Invariant of while. Flexible version to be used everywhere *)
Definition invPrime :
   Store gPr -> Z -> bool -> Prop
:=
   fun s : Store gPr =>
   fun i : Z =>
   fun pr : bool =>
      Zmod i 2 <> 0 /\ s n >= 3 /\ i >= 3 /\ i <= s n /\
      (pr = true <-> Zmod (s n) 2 <> 0 /\
                     forall m : Z, Zmod m 2 <> 0 -> 3 <= m -> m < i
                                   -> Zmod (s n) m <> 0)
.

Lemma pPrimePre :
pred_implies gPr
   (fun s : Store gPr => s n >= 3)
   (fun s : Store gPr => invPrime s 3 (Zneq_bool (Zmod (s n) 2) 0))
.

Proof.
unfold pred_implies in |- *.
unfold invPrime in |- *.
intros s H.
constructor.
 simplify_eq.
 constructor.
  assumption.
  constructor.
   auto with zarith.
   constructor.
    auto with zarith.
    split.
     intro H0.
       constructor.
      apply Zne_bool_is_ne.
        assumption.
      intros.
        absurd (3 <= m).
       apply Zgt_not_le.
         apply Zlt_gt.
         assumption.
       assumption.
     intros.
       destruct H0.
       apply Zne_is_ne_bool.
       assumption.
Qed.

Lemma pPrimeWhile2 :
pred_implies gPr
   (fun s : Store gPr => invPrime s (s i) (s pr) /\
      (s i) * (s i) <= (s n) /\ (s pr) = true)
   (fun s : Store gPr => invPrime s (s i + 2)
                                    (Zneq_bool (Zmod (s n) (s i)) 0))
.

Proof.
unfold pred_implies in |- *.
unfold invPrime in |- *.
intros s H.
destruct H as (H1, (Hisq_le_n, Hpr)).
destruct H1 as (Hi_m_2, (Hn_ge_3, (Hi_3, (Hi_n, H)))).
constructor.
 assert ((s i + 2) mod 2 = s i mod 2).
  apply Z_mod_plus with (c := 2) (b := 1) (a := s i).
    auto with zarith.
  rewrite H0.
    assumption.
 constructor.
  assumption.
  constructor.
   auto with zarith.
   constructor.
    apply Zle_trans with (m := s i * s i).
     apply Zplus_le_reg_r with (p := - s i).
       replace (s i + 2 + - s i) with 2.
      replace (s i * s i + - s i) with (s i * (s i - 1)).
       apply Zle_trans with 6.
        auto with zarith.
        change 6 with (3 * 2) in |- *.
          apply Zmult_le_compat.
         auto with zarith.
         auto with zarith.
         auto with zarith.
         auto with zarith.
       change (s i - 1) with (s i + -1) in |- *.
         replace (- s i) with (s i * -1).
        auto with zarith.
        auto with zarith.
      auto with zarith.
     assumption.
    destruct H.
      assert
    (s n mod 2 <> 0 /\
        (forall m : Z, m mod 2 <> 0 -> 3 <= m -> m < s i -> s n mod m <> 0)).
     rewrite Hpr in H.
       apply H.
       trivial.
     clear H H0.
       destruct H1 as (Hn_m_2, Hmi).
       split.
      intro.
        assert (s n mod s i <> 0).
       apply Zne_bool_is_ne.
         assumption.
       clear H.
         constructor.
        assumption.
        clear Hpr.
          intros.
          assert (Zsucc (Zsucc (s i)) > m).
         unfold Zsucc in |- *.
           apply Zlt_gt.
           replace (s i + 2) with (s i + 1 + 1) in H2.
          assumption.
          auto with zarith.
         clear H2.
           assert (Zsucc (s i) > m \/ m = Zsucc (s i)).
          apply Zgt_succ_gt_or_eq.
            assumption.
          assert (m = Zsucc (s i) \/ s i > m \/ m = s i).
           destruct H2.
            right.
              apply Zgt_succ_gt_or_eq.
              assumption.
            left; assumption.
           clear H2 H3.
             destruct H4.
            absurd (m mod 2 <> 0).
             rewrite H2.
               unfold not in |- *.
               intro.
               elim H3.
               apply Zmod_2_succ.
              auto with zarith.
              assumption.
             assumption.
            destruct H2.
             apply Hmi.
              assumption.
              assumption.
              apply Zgt_lt.
                assumption.
             rewrite H2.
               assumption.
      intros.
        destruct H.
        clear H.
        apply Zne_is_ne_bool.
        apply H0.
       assumption.
       apply Zge_le.
         assumption.
       auto with zarith.
Qed.

Lemma pPrimeCond :
pred_implies gPr
            (fun s : Store gPr => invPrime s (s i) (s pr))
            (fun s : Store gPr => ( if andb (Zle_bool ((s i) * (s i)) (s n))
                                            (s pr)
                                    then
                                       (fun s : Store gPr =>
                                          invPrime s (s i) (s pr) /\
                                          (s i) * (s i) <= (s n) /\
                                          (s pr) = true)
                                    else
                                       (fun s : Store gPr =>
                                          invPrime s (s i) (s pr) /\
                                          ~ ((s i) * (s i) <= (s n) /\
                                             (s pr) = true)) ) s)
.

Proof.
unfold pred_implies in |- *.
intros.
destruct (sumbool_of_bool (Zle_bool (s i * s i) (s n) && s pr)); rewrite e.
 constructor.
  assumption.
  unfold Zle_bool in e.
    unfold Zle in |- *.
    induction (s i * s i ?= s n).
   constructor; [ simplify_eq | assumption ].
   constructor; [ simplify_eq | assumption ].
   discriminate e.
 constructor.
  assumption.
  apply or_not_and.
    unfold Zle_bool in e.
    unfold Zle in |- *.
    induction (s i * s i ?= s n).
   right.
     change (true && s pr = false) with (s pr = false) in e.
     rewrite e.
     auto with zarith.
   change (true && s pr = false) with (s pr = false) in e.
     right; rewrite e.
     auto with zarith.
   left.
     auto with zarith.
Qed.


Lemma pPrimePost:
pred_implies gPr
   (fun s : Store gPr => invPrime s (s i) (s pr) /\
                         ~ ((s i) * (s i) <= (s n) /\ (s pr) = true))
   (fun s : Store gPr => s pr = true <-> Zprime (s n))
.

Proof.
unfold pred_implies in |- *.
unfold invPrime in |- *.
intros s H.
destruct H as (H1, Hs).
destruct H1 as (Hi_mod2, (Hn_gt3, (Hi_gt3, (Hi_n, H)))).
assert (Hssqr : ~ s i * s i <= s n \/ s pr <> true).
 apply not_and_or.
   assumption.
 clear Hs.
   split.
  destruct H.
    intro Hpr.
    assert
     (s n mod 2 <> 0 /\
      (forall m : Z, m mod 2 <> 0 -> 3 <= m -> m < s i -> s n mod m <> 0)).
   rewrite Hpr in H.
     apply H.
     trivial.
   clear H H0.
     destruct H1 as (Hn_mod2, H).
     destruct Hssqr.
    apply Zprime_sqrt_mod2.
     assumption.
     intros.
       apply H.
      assumption.
      assumption.
      apply Zlt_square_simpl.
       auto with zarith.
       auto with zarith.
    absurd (s pr <> true).
     rewrite Hpr.
       auto with zarith.
     assumption.
  unfold Zprime in |- *.
    intro.
    destruct H.
    apply H1.
    clear H H1.
    constructor.
   apply H0.
     auto with zarith.
   intros.
     apply H0.
     constructor.
    auto with zarith.
    apply Zlt_le_trans with (m := s i).
     assumption.
     assumption.
Qed.

Definition pPrime :=
   acseq
      (acseq
         (acassert gPr
            (fun s : Store gPr => s n >= 3)
            (fun s : Store gPr => invPrime s 3 (Zneq_bool (Zmod (s n) 2) 0))
            pPrimePre
         )
         (acseq
            (acassign pr (aebinop opNe (aebinop opMod (aevar n) (aeint 2))
                                       (aeint 0)
                         )
            )
            (acassign i (aeint 3))
         )
      )
      (acseq
         (acwhile gPr
            (fun s : Store gPr => invPrime s (s i) (s pr))
            (aeweak gPr
               (fun s : Store gPr => invPrime s (s i) (s pr))
               (fun s : Store gPr => (if andb (Zle_bool ((s i) * (s i)) (s n))
                                               (s pr)
                                       then
                                          (fun s : Store gPr =>
                                             invPrime s (s i) (s pr) /\
                                             (s i) * (s i) <= (s n) /\
                                             (s pr) = true)
                                       else
                                          (fun s : Store gPr =>
                                             invPrime s (s i) (s pr) /\
                                             ~ ((s i) * (s i) <= (s n) /\
                                                (s pr) = true)) ) s)
               pPrimeCond
               (aebinop opAnd
                  (aebinop opLe (aebinop opTimes (aevar i) (aevar i)) (aevar n))
                  (aevar pr)
               )
            ) 
            (acseq
               (acseq
                  (acassert gPr
                     (fun s : Store gPr => invPrime s (s i) (s pr) /\
                               (s i) * (s i) <= (s n) /\ (s pr) = true)
                     (fun s : Store gPr => invPrime s (s i + 2)
                                             (Zneq_bool (Zmod (s n) (s i)) 0))
                  pPrimeWhile2
                  )
                  (acassign pr (aebinop opNe (aebinop opMod (aevar n) (aevar i))
                                             (aeint 0)
                               )
                  )
                  (*(acif (aebinop opEq (aebinop opMod (aevar n) (aevar i))
                                        (aeint 0))
                     (acassign pr (aebool false))
                     (acskip)
                  )*)
               )
               (acassign i (aebinop opPlus (aevar i) (aeint 2)))
            )
         )
         (acassert gPr
            (fun s : Store gPr => invPrime s (s i) (s pr) /\
                                  ~ ((s i) * (s i) <= (s n) /\ (s pr) = true))
            (fun s : Store gPr => s pr = true <-> Zprime (s n))
            pPrimePost
         )
      )
. 
 
Theorem dPrime :
   typAC gPr pPrime.

Proof.
typeAC.
Defined.

Lemma wpPrime :
   wpC gPr pPrime dPrime
      (fun s : Store gPr => s pr = true <-> Zprime (s n))
      = (fun s: Store gPr => s n >= 3)
.

Proof.
reflexivity.
Qed.

Theorem sPrime :
   SpecAC gPr pPrime
      (fun s: Store gPr => s n >= 3)
      (fun s : Store gPr => s pr = true <-> Zprime (s n))
.

Proof.
specAC dPrime.
Qed.

End ExamplePrimality.
