(*
 * Hoare Logic,
 * Annotated version --> Logic
 * 
 * CVS : $Id: tlhlAnnotLogic.v,v 1.5 2005/04/12 01:29:56 nickie Exp $
*)

Require Export useful.
Require Export tlhlAnnotDef.
Require Export tlhlAnnotTyp.
Require Export tlhlAnnotSem.

Inductive SpecAE (g : Env) :
   forall e : AExpr, forall t : Omega, SPECE g t
:=
    axiomAEint :
       forall n : Int,
       forall F : Int -> Pred g,
         SpecAE g (aeint n) uint (F n) F
  | axiomAEbool :
       forall b : Bool,
       forall F : Bool -> Pred g,
         SpecAE g (aebool b) ubool (F b) F
  | axiomAEvar :
       forall x : Var,
       forall F : semT (g x) -> Pred g,
         SpecAE g (aevar x) (g x) (fun s : Store g => F (s x) s) F
  | axiomAEunop :
       forall op : UnOp,
       forall e  : AExpr,
       forall P  : Pred g,
       forall F  : semT (unopTR op) -> Pred g,
       SpecAE g e (unopT1 op) P
           (fun v : semT (unopT1 op) => F (unopFun op v)) ->
         SpecAE g (aeunop op e) (unopTR op) P F
  | axiomAEbinop :
      forall op : BinOp,
      forall e1 : AExpr,
      forall e2 : AExpr,
      forall P  : Pred g,
      forall F  : semT (binopTR op) -> Pred g,
      forall G  : semT (binopT1 op) -> Pred g,
      SpecAE g e1 (binopT1 op) P G ->
      (forall v1 : semT (binopT1 op),
              SpecAE g e2 (binopT2 op) (G v1)
	        (fun v2 : semT (binopT2 op) => F (binopFun op v1 v2))) ->
        SpecAE g (aebinop op e1 e2) (binopTR op) P F
  | axiomAEweak :
      forall P  : Pred g,
      forall Q  : Pred g,
      forall p  : pred_implies g P Q,
      forall e  : AExpr,
      forall t  : Omega, 
      forall F  : semT t -> Pred g,
        SpecAE g e t Q F ->
           SpecAE g (aeweak g P Q p e) t P F
.

Inductive SpecAC (g : Env) :
   forall c : AComm, SPECC g
:=
   axiomACskip :
     forall P : Pred g,
       SpecAC g acskip P P
 | axiomACassign :
     forall x  : Var,
     forall e  : AExpr,
     forall P  : Pred g,
     forall Q  : Pred g,
     SpecAE g e (g x) P
         (fun v : semT (g x) => fun s : Store g => Q (updateStore g x v s)) ->
       SpecAC g (acassign x e) P Q
 | axiomACseq :
     forall c1 : AComm,
     forall c2 : AComm,
     forall P  : Pred g,
     forall Q  : Pred g,
     forall R  : Pred g,
     SpecAC g c1 P R ->
     SpecAC g c2 R Q ->
       SpecAC g (acseq c1 c2) P Q
 | axiomACif :
     forall e  : AExpr,
     forall c1 : AComm,
     forall c2 : AComm,
     forall P  : Pred g,
     forall Q  : Pred g,
     forall F  : semT ubool -> Pred g,
     SpecAE g e ubool P F ->
     SpecAC g c1 (F true) Q ->
     SpecAC g c2 (F false) Q ->
       SpecAC g (acif e c1 c2) P Q
 | axiomACwhile :
     forall P  : Pred g,
     forall e  : AExpr,
     forall c  : AComm,
     forall F : semT ubool -> Pred g,
     SpecAE g e ubool P F ->
     SpecAC g c (F true) P ->
       SpecAC g (acwhile g P e c) P (F false)
 | axiomACassert :
      forall P : Pred g,
      forall Q : Pred g,
      forall p : pred_implies g P Q,
        SpecAC g (acassert g P Q p)P Q
.

(* use sparingly... *)
Axiom ext :
   forall g : Env,
   forall t : Omega,
     extensionality (semT t) (fun _ : semT t => Pred g)
.

Theorem SpecEisUnique :
   forall g : Env,
   forall e : AExpr,
   forall t : Omega,
   forall P : Pred g,
   forall F : semT t -> Pred g,
     SpecAE g e t P F ->
     forall Q : Pred g,
       SpecAE g e t Q F ->
         pred_equivalent g P Q
.
Proof.
unfold pred_equivalent in |- *.
intros g e.
induction e.
(*int*)
 intros t P F Hspec1 Q Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   replace F1 with F0.
  reflexivity.
  apply inj_pairT2 with (P := fun t : Omega => semT t -> Pred g) (p := uint).
    transitivity (existT (fun t : Omega => semT t -> Pred g) t F).
   assumption.
   symmetry  in |- *.
     assumption.
(*bool*)
 intros t P F Hspec1 Q Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   replace F1 with F0.
  reflexivity.
  apply inj_pairT2 with (P := fun t : Omega => semT t -> Pred g) (p := ubool).
    transitivity (existT (fun t : Omega => semT t -> Pred g) t F).
   assumption.
   symmetry  in |- *.
     assumption.
(*var*)
 intros t P F Hspec1 Q Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   replace F1 with F0.
  reflexivity.
  apply inj_pairT2 with (P := fun t : Omega => semT t -> Pred g) (p := g v).
    transitivity (existT (fun t : Omega => semT t -> Pred g) t F).
   assumption.
   symmetry  in |- *.
     assumption.
(*unop*)
 intros t P F Hspec1 Q Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   apply IHe with (F := fun v : semT (unopT1 u) => F0 (unopFun u v)).
  assumption.
  replace F0 with F1.
   assumption.
   apply inj_pairT2 with
       (P := fun t : Omega => semT t -> Pred g) (p := unopTR u).
     transitivity (existT (fun x : Omega => semT x -> Pred g) t F).
    assumption.
    symmetry  in |- *.
      assumption.
(*binop*)
 intros t P F Hspec1 Q Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   apply IHe1 with (F := G).
  assumption.
  replace G with G0.
   assumption.
   apply ext.
     intro v1.
     apply IHe2 with
         (F := fun v2 : semT (binopT2 b) => F1 (binopFun b v1 v2)).
    apply H16.
    replace F1 with F0.
     apply H7.
     apply inj_pairT2 with
         (P := fun t : Omega => semT t -> Pred g) (p := binopTR b).
       transitivity (existT (fun x : Omega => semT x -> Pred g) t F).
      assumption.
      symmetry  in |- *.
        assumption.
(*weak*)
 intros t P' F Hspec1 Q' Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   rewrite <- H4.
   replace P1 with P3.
  assumption.
  apply inj_pairT2 with (P := Pred).
    transitivity (existT (fun g : Env => Pred g) g0 P).
   assumption.
   symmetry  in |- *.
     assumption.
Qed.

Theorem SpecCisUnique :
   forall g : Env,
   forall c : AComm,
   forall P : Pred g,
   forall R : Pred g,
     SpecAC g c P R ->
     forall Q : Pred g,
       SpecAC g c Q R ->
         pred_equivalent g P Q
.
Proof.
unfold pred_equivalent in |- *.
intros g c.
induction c.
(*skip*)
 intros P R Hspec1 Q Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   reflexivity.
(*assign*)
 intros P R Hspec1 Q Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   apply SpecEisUnique with
       (g := g)
       (e := e)
       (t := g v)
       (P := P)
       (Q := Q)
       (F := fun (v0 : semT (g v)) (s : Store g) => R (updateStore g v v0 s)).
  rewrite H1.
    assumption.
  rewrite H1.
    assumption.
(*seq*)
 intros P R Hspec1 Q Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   apply IHc1 with (R := R0).
  assumption.
  replace R0 with R1.
   assumption.
   apply IHc2 with (R := R).
    assumption.
    assumption.
(*if*)
 intros P R Hspec1 Q Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   apply SpecEisUnique with
       (g := g)
       (e := e)
       (t := ubool)
       (P := P)
       (Q := Q)
       (F := F).
  rewrite H.
    assumption.
  replace F with F0.
   rewrite H.
     assumption.
   apply ext.
     intro b.
     induction b.
    apply IHc1 with (R := R).
     assumption.
     assumption.
    apply IHc2 with (R := R).
     assumption.
     assumption.
(*while*)
 intros P R Hspec1 Q Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   rewrite <- H1.
   replace P1 with P2.
  assumption.
  apply inj_pairT2 with (P := Pred).
    transitivity (existT (fun g : Env => Pred g) g0 p).
   assumption.
   symmetry  in |- *.
     assumption.
(*assert*)
 intros P' R Hspec1 Q' Hspec2.
   inversion Hspec1.
   inversion Hspec2.
   rewrite <- H3.
   replace P1 with P3.
  assumption.
  apply inj_pairT2 with (P := Pred).
    transitivity (existT (fun g : Env => Pred g) g0 P).
   assumption.
   symmetry  in |- *.
     assumption.
Qed.
