(*
 * Hoare Logic,
 * Annotated version --> Semantics
 * 
 * CVS : $Id: tlhlAnnotSem.v,v 1.4 2005/03/24 16:45:31 nickie Exp $
*)

Require Export tlhlAnnotDef.

Inductive semAE (g : Env) :
   forall e : AExpr,
   forall t : Omega,
      Store g -> semT t -> Type
:=
   semAEint :
      forall n : Int,
      forall s : Store g,
         semAE g (aeint n) uint s n
 | semAEbool :
      forall b : Bool,
      forall s : Store g,
         semAE g (aebool b) ubool s b
 | semAEvar :
      forall x : Var,
      forall s : Store g,
         semAE g (aevar x) (g x) s (s x)
 | semAEunop :
      forall op : UnOp,
      forall e  : AExpr,
      forall s  : Store g,
      forall v  : semT (unopT1 op),
      semAE g e (unopT1 op) s v ->
         semAE g (aeunop op e) (unopTR op) s (unopFun op v)
 | semAEbinop :
      forall op : BinOp,
      forall e1 : AExpr,
      forall e2 : AExpr,
      forall s  : Store g,
      forall v1 : semT (binopT1 op),
      forall v2 : semT (binopT2 op),
      semAE g e1 (binopT1 op) s v1 ->
      semAE g e2 (binopT2 op) s v2 ->
         semAE g (aebinop op e1 e2) (binopTR op) s (binopFun op v1 v2)
 | semAEweak :
      forall P : Pred g,
      forall Q : Pred g,
      forall p : pred_implies g P Q,
      forall e : AExpr,
      forall t : Omega,
      forall s : Store g,
      forall v : semT t,
         semAE g e t s v -> semAE g (aeweak g P Q p e) t s v
.

Inductive semAC (g : Env) :
   forall c : AComm,
      Store g -> Store g -> Type
:=
   semACskip :
      forall s : Store g,
         semAC g acskip s s
 | semACassign :
      forall x  : Var,
      forall e  : AExpr,
      forall s : Store g,
      forall v : semT (g x),
      semAE g e (g x) s v ->
         semAC g (acassign x e)
              s (updateStore g x v s)
 | semACseq :
      forall c1 : AComm,
      forall c2 : AComm,
      forall s  : Store g,
      forall s1 : Store g,
      forall s2 : Store g,
      semAC g c1 s s1 ->
      semAC g c2 s1 s2 ->
         semAC g (acseq c1 c2) s s2
 | semACift :
      forall e  : AExpr,
      forall c1 : AComm,
      forall c2 : AComm,
      forall s  : Store g,
      forall s1 : Store g,
      semAE g e ubool s true ->
      semAC g c1 s s1 ->
         semAC g (acif e c1 c2) s s1
 | semACiff :
      forall e  : AExpr,
      forall c1 : AComm,
      forall c2 : AComm,
      forall s  : Store g,
      forall s2 : Store g,
      semAE g e ubool s false ->
      semAC g c2 s s2 ->
         semAC g (acif e c1 c2) s s2
 | semACwhilet :
      forall P  : Pred g,
      forall e  : AExpr,
      forall c  : AComm,
      forall s  : Store g,
      forall s1 : Store g,
      forall s2 : Store g,
      semAE g e ubool s true ->
      semAC g c s s1 ->
      semAC g (acwhile g P e c) s1 s2 ->
         semAC g (acwhile g P e c) s s2
 | semACwhilef :
      forall P  : Pred g,
      forall e  : AExpr,
      forall c  : AComm,
      forall s  : Store g,
      semAE g e ubool s false ->
         semAC g (acwhile g P e c) s s
 | semACassert :
      forall P  : Pred g,
      forall Q  : Pred g,
      forall p  : pred_implies g P Q,
      forall s  : Store g,
         semAC g (acassert g P Q p) s s
.
