(*
 * Hoare Logic,
 * Annotated version --> Soundness
 * 
 * CVS : $Id: tlhlAnnotSound.v,v 1.3 2005/04/12 01:29:56 nickie Exp $
*)

Require Export tlhlAnnotSem.
Require Export tlhlAnnotLogic.

Theorem LogicIsSoundAE:
   forall g : Env,
   forall e : AExpr,
   forall t : Omega,
   forall d : typAE g e t,
   forall P : Pred g,
   forall F : semT t -> Pred g,
   SpecAE g e t P F ->
   forall s : Store g,
   forall v : semT t,
   semAE g e t s v ->
      P s -> F v s
.
Proof.
intros g e t d P F Hspec.
induction Hspec.
 intros s v Hsem.
   inversion Hsem.
   assert (n = v).
  apply inj_pair2 with (P := semT) (p := uint).
    assumption.
  rewrite H.
    trivial.
 intros s v Hsem.
   inversion Hsem.
   assert (b = v).
  apply inj_pair2 with (P := semT) (p := ubool).
    assumption.
  rewrite H.
    trivial.
 intros s v Hsem.
   inversion Hsem.
   assert (s x = v).
  apply inj_pair2 with (P := semT) (p := g x).
    assumption.
  rewrite H2.
    trivial.
 intros s v Hsem.
   inversion Hsem.
   assert (unopFun op v0 = v).
  apply inj_pair2 with (P := semT) (p := unopTR op).
    assumption.
  rewrite <- H3.
    apply IHHspec.
   inversion d.
   assumption.
   assumption.
 intros s v Hsem.
   inversion Hsem.
   assert (binopFun op v1 v2 = v).
  apply inj_pair2 with (P := semT) (p := binopTR op).
    assumption.
  rewrite <- H5.
    intro Hpre.
    apply H0.
   inversion d.
     assumption.
   assumption.
   apply IHHspec.
    inversion d.
      assumption.
    assumption.
    assumption.
 intros s v Hsem.
   inversion Hsem.
   intro Hpre.
   apply IHHspec.
  assert (v0 = v).
   apply inj_pair2 with (P := semT) (p := t).
     assumption.
   inversion d.
     assumption.
  assert (v0 = v).
   apply inj_pair2 with (P := semT) (p := t).
     assumption.
   rewrite <- H5.
     assumption.
  apply p.
    assumption.
Qed.

Lemma while_ind :
   forall g : Env,
   forall I : Pred g,
   forall e : AExpr,
   forall c : AComm,
   forall P : Store g -> Store g -> Prop,
      (forall s : Store g,
         semAE g e ubool s false -> P s s) ->
      (forall s  : Store g,
       forall s' : Store g,
       forall s'' : Store g,
         semAE g e ubool s true ->
         semAC g c s s' ->
         semAC g (acwhile g I e c) s' s'' ->
            P s' s'' -> P s s'') ->
      forall s  : Store g,
      forall s' : Store g,
         semAC g (acwhile g I e c) s s' -> P s s'
.
Proof.
intros g I e c P Hf Ht s s'.
set (cw := acwhile g I e c) in |- *.
assert (cw = acwhile g I e c).
 unfold cw in |- *.
   reflexivity.
 intro Hsem.
   induction Hsem.
  discriminate H.
  discriminate H.
  discriminate H.
  discriminate H.
  discriminate H.
  injection H.
    intros H1 H2 H3.
    apply Ht with (s' := s1).
   rewrite <- H2.
     assumption.
   rewrite <- H1.
     assumption.
   rewrite <- H1.
     rewrite <- H2.
     replace I with P0.
    assumption.
    apply inj_pairT2 with (P := Pred).
      assumption.
   apply IHHsem2.
     assumption.
  apply Hf.
    injection H.
    intros H1 H2 H3.
    rewrite <- H2.
    assumption.
  discriminate H.
Qed.

Theorem LogicIsSoundC:
   forall g : Env,
   forall c : AComm,
   forall d : typAC g c,
   forall P : Pred g,
   forall Q : Pred g,
   SpecAC g c P Q ->
   forall s : Store g,
   forall s' : Store g,
   semAC g c s s' ->
      P s -> Q s'
.
Proof.
intros g c d P Q Hspec.
induction Hspec.
 intros s s' Hsem.
   inversion Hsem.
   trivial.
 intros s s' Hsem.
   inversion Hsem.
   inversion d.
   intro Hpre.
   apply LogicIsSoundAE with
       (e := e)
       (t := g x)
       (P := P)
       (F := fun (v : semT (g x)) (s : Store g) => Q (updateStore g x v s)).
  assumption.
  assumption.
  assumption.
  assumption.
 intros s s' Hsem.
   inversion Hsem.
   inversion d.
   intro Hpre.
   apply IHHspec2 with (s := s1).
  assumption.
  assumption.
  apply IHHspec1 with (s := s).
   assumption.
   assumption.
   assumption.
 intros s s' Hsem.
   inversion Hsem.
  inversion d.
    intro Hpre.
    apply IHHspec1 with (s := s).
   assumption.
   assumption.
   apply LogicIsSoundAE with (e := e) (t := ubool) (P := P) (F := F).
    assumption.
    assumption.
    assumption.
    assumption.
  inversion d.
    intro Hpre.
    apply IHHspec2 with (s := s).
   assumption.
   assumption.
   apply LogicIsSoundAE with (e := e) (t := ubool) (P := P) (F := F).
    assumption.
    assumption.
    assumption.
    assumption.
 intros s s' Hsem.
   inversion d.
   apply while_ind with
       (I := P)
       (e := e)
       (c := c)
       (P := fun s s' : Store g => P s -> F false s').
  intros se Hseme Hpre.
    apply LogicIsSoundAE with (e := e) (P := P) (F := F).
   assumption.
   assumption.
   assumption.
   assumption.
  intros se sc sf Hseme Hsemc Hsem' HI Hpre.
    apply HI.
    apply IHHspec with (s := se).
   assumption.
   assumption.
   apply LogicIsSoundAE with (e := e) (P := P) (F := F).
    assumption.
    assumption.
    assumption.
    assumption.
  assumption.
 intros s s' Hsem.
   inversion Hsem.
   inversion d.
   intro Hpre.
   apply p.
   assumption.
Qed.
